import datetime
import logging
from pathlib import Path
from typing import List, Union

import box
import torch
import torch.nn as nn
from torch.optim import Optimizer

from .configs import StateFile


def save_state_file(state: StateFile, file_path: Union[str, Path]):
    Path(file_path).parent.mkdir(exist_ok=True, parents=True)
    torch.save(vars(state), str(file_path))


def load_state_file(file_path: Union[str, Path], device: str = 'cuda') -> StateFile:
    checkpoint = torch.load(str(file_path),
                            map_location=(lambda tensor, _: tensor.cuda()) if device == 'cuda' else device)
    return StateFile(checkpoint)


def to_state(model: nn.Module, config: dict = None, optimizers: List[Optimizer] = (), info: dict = None) -> StateFile:
    state = StateFile()
    state.model = model.state_dict()
    state.config = config
    state.optimizers = [optimizer.state_dict() for optimizer in optimizers]
    state.info = info
    state.timestamp = str(datetime.datetime.now())
    return state


def from_state(state: StateFile, model: nn.Module = None, optimizers: List[Optimizer] = (),
               device: str = 'cuda', strict: bool = True) -> nn.Module:
    if model is None:
        model: nn.Module = box.factory(state.config, tag='model').to(device)
    model.load_state_dict(state.model, strict=strict)
    if not strict:  # pragma: no cover
        model_keys = set(model.state_dict().keys())
        saved_keys = state.model.keys()
        missing_keys = model_keys - saved_keys
        if missing_keys:
            logging.warning('caution: model state missing keys {}'.format(list(missing_keys)))
    for optimizer, state_dict in zip(optimizers, state.optimizers):
        optimizer.load_state_dict(state_dict)
    return model


def load_model_from_state(file_path: Union[str, Path], device: str = 'cuda') -> nn.Module:
    state = load_state_file(file_path=file_path, device=device)
    return from_state(state=state, device=device)



import json
from pathlib import Path

import argparse_schema
import torch.onnx

from .configs import ArgumentConfig
from .state_utils import load_state_file, from_state


def cli_main():  # pragma: no cover
    with open(str(Path(__file__).parent / 'schema' / 'argument_config.json')) as f:
        schema = json.load(f)
    argument = ArgumentConfig(argparse_schema.parse(schema))

    state = load_state_file(argument.state_file, device=argument.device)

    if argument.extra_import:
        __import__(argument.extra_import)
    model = from_state(state, device=argument.device) if argument.load_model or argument.export_onnx else None

    del state.model
    del state.optimizers

    print(json.dumps(vars(state), indent=2))
    if argument.load_model:
        print(model)

    if argument.export_onnx:
        dummy_input = torch.randn(1, 3, argument.input_size, argument.input_size)
        torch.onnx.export(model, dummy_input, argument.export_onnx, verbose=argument.verbose)
